""" Hark Base Node is a base class for PyCodeExecutor in hark-python.

* Introduction
You can take advantage of Python modules for HARK. 
For example, 
you can plot using pylab or matplotlib as you want, 
you can write a small test code without having any source codes.



* How to use
To implement your python code for PyCodeExecutor, 
you need to do following things.
  (1) define a class that inherits HarkBaseNode. 
      The default class name is HarkNode. 
      (you can specify any name from the parameer of PyCodeExecutor.)
  (2) in __init__ method, 
        set the tuple self.outputNames to define the output terminal names, and 
        set the tuple self.outputTypes to define the output terminal types.
  (3) in calculate method,
        get input values from instance variables,
        implement your code in calculate method, and
        set your calculation results to the dictionary self.outputValues.
Detailed descriptions are written in each method.



* Sample code and network
This is the simplest sample.
other sample scripts and network files are in example/ directory.
run the shell script run_examples.sh to run them.

** Network
   Add the PyCodeExecutor node,
   add input1, input2, sum, and sqrt by hand.
                               |--------------|
       [Constant: int: 5] -----|input1     sum|---[Print]=OUTPUT
       [Cnostant: float: 2.5]--|input2    sqrt|---[Print]=OUTPUT_1
                               |--------------|

**Code
import harkbasenode  # This is the class defined here.
import numpy         # for calculate square root.

class HarkNode(harkbasenode.HarkBaseNode):   # inherit
    def __init__(self):
        # This node has two outputs: 
        # sum of primitive float, and sqrt of primitive complex.
        self.outputNames = ("sum", "sqrt")
        self.outputTypes = ("prim_float", "prim_complex)

    def calculate(self):
        # for each frame, these two values are calculated.
        self.outputValues["sum"] = self.input1 + self.input2
        self.outputValues["sqrt"] = numpy.sqrt(self.output1 * self.output2)



* Advanced tips
(1) Additional instance variables
You have some system variables such as 
  self.count: 
     The current number of frame. 
     You can "plot for each 50 frames" using this value.
  self.nodeName
     A node's name which is unique in the network, e.g., node_PyCodeExecutor_1.
  self.nodeID
     A node's number which is unique in the network, 
     e.g., nodeID = 1 if the node name is node_PyCodeExecutor_1.
     You can plot in different figure using:
         pylab.figure(self.nodeID)

"""

__author__ = "Takeshi Mizumoto"
__date__ = "25 Oct. 2011"

import numpy
import exceptions    

class HarkBaseNode:
    """A base class for PyCodeExecutor in hark-python."""
    def __init__(self):
        """In this constructor, you need to specify output names and types as tuples.
e.g., 
    self.outputNames = ("output1", "output2")
    self.outputTypes = ("prim_float", "matrix_complex")

You can use [a-zA-Z0-9_] for names, and 
you can use [prim|vector|matrix]_[int|float|complex] for types.
If you specified wrong type, TypeError exception will be raised.
"""
        self.outputNames = ()
        self.outputTypes = ()
        self.outputValues = {}

        raise exceptions.NotImplementedError("You must overload the constructor")

    def calculate(self):
        """For each iteration in flowdesigner, this function is called.

[Getting input]
Before the call, PyCodeExecutor gives the input values as instance variables.
e.g., 
Suppose tht your PyCodeExecutor node is like this:
         |---------------|
     ----|fft      output|---
     ----|integer        |
         |---------------|
Then, you can get the fft and input as:
   self.fft
   self.integer
to know all the inputs, use the following code:
   print dir()


[Setting output]
Set the output values to the dictionary self.outputValues. 
Its keys are self.outputNames defined in __init__. 
"""

        pass

    def makeup(self):
        """ Internal method. 
You do not need run this method. This will automatically called in PyCodeExecutor.
This builds the dictionary to give the output values to flowdesigner, 
and checks type strings."""
        
        if len(self.outputTypes) != len(self.outputNames):
            raise exceptions.ValueError(
                "output Names and types do not match: " + \
                    "types = " + str(len(self.outputTypes)) + \
                    "names = " + str(len(self.outputNames)))
                

        for otype in self.outputTypes:
            if not(otype.split("_")[0] in ["prim", "vector", "matrix", "map"]):
                raise exceptions.TypeError(
                    "outputTypes must be prim_*, vector_*, or matrix_*. "
                    + "you specified: " + otype)
            if not(otype.split("_")[-1] in ["int", "float", "complex", "source"]):
                raise exceptions.TypeError(
                    "outputTypes must be *_int, *_float, *_complex, or *_source"
                    + "you specified: " + otype)
            
        self.outputValues = dict(zip(self.outputNames,
                                     [None for i in range(len(self.outputNames))]))
        self.outputNames = tuple(self.outputNames)
        self.outputTypes = tuple(self.outputTypes)

    def convertValues(self):
        """ Internal method.
You do not need run this method. This will automatically called in PyCodeExecutor.
This method converts a python-friendly types into C++-friendly types, 
such as from numpy.array to vector<int>."""

        for outname in self.outputNames:
            outtype = self.outputValues[outname]
            # if None, raise exception
            if outtype == None:
                raise exceptions.ValueError(outname + " is not assigned yet.")

        for ioutname, outname in enumerate(self.outputNames):
            outtype = self.outputTypes[ioutname]

            # decide type
            if "int" in outtype: dtype = numpy.int32
            elif "float" in outtype: dtype = numpy.float32
            elif "complex" in outtype: dtype = numpy.complex64
            elif "source" in outtype: dtype = "source"
            else:
                raise exceptions.ValueError("unknown type: " + outtype)

            # convert to given type
            if outtype.startswith("prim"):
                if "source" != dtype:
                    self.outputValues[outname] =\
                        self.__convertType(self.outputValues[outname], dtype)
            elif outtype.startswith("map"):
                for key in self.outputValues[outname].keys():
                    item = self.outputValues[outname][key]
                    if hasattr(item, "__len__"):
                        self.outputValues[outname][key]= list(item)
                    else:
                        self.outputValues[outname][key] = self.__convertType(item, dtype)
            elif outtype.startswith("vector"):
                if "source" != dtype:
                    self.outputValues[outname] \
                        = numpy.array(self.outputValues[outname], dtype).tolist()
            elif outtype.startswith("matrix"):
                self.outputValues[outname] \
                    = numpy.array(self.outputValues[outname], dtype).tolist()
            else:
                raise exceptions.ValueError("unknown type: " + outtype)

    def __convertType(self, value, dtype):
        """Internal method for convert Values"""
        if dtype == numpy.float32: return float(value)
        elif dtype == numpy.int32: return int(value)
        elif dtype == numpy.complex64: return complex(value)
        elif dtype == "source": return value

    def __del__(self):
        pass
